from ruamel.yaml import YAML
from ..types import ID, Attachment


def get_yaml():
    yaml = YAML(typ='safe')
    yaml.width = 120
    yaml.default_flow_style = False
    yaml.register_class(ID)
    yaml.register_class(Attachment)
    return yaml



from haul.policy import ImportPolicy, RelinkAction
from io import BytesIO
from test_app.app import reset
from test_app.models import Book, Author, Tag
from test_app.export import BookExporter, AuthorExporter, TagExporter

from haul import ImportContainer, ExportContainer


EXPORTERS = [BookExporter, AuthorExporter, TagExporter]


def test_report_discard():
    reset()
    a1 = Author.objects.create(name='1')
    Author.objects.create(name='2')
    Book.objects.create(name='b1', author=a1)

    d = BytesIO()

    c = ExportContainer(exporters=EXPORTERS)
    c.export_objects(Author.objects.all())
    c.export_objects(Book.objects.all())
    c.write(d)

    reset()

    class Policy(ImportPolicy):
        def relink_object(self, model_cls, obj):
            if model_cls is Author and obj.fields['name'] == '1':
                return RelinkAction.Discard()
            return RelinkAction.Create()

    c = ImportContainer(exporters=EXPORTERS, policy=Policy())
    with c.read(d):
        report = c.import_objects()

    assert report.imported_objects == set(Author.objects.all())
    assert len(report.discarded_objects) == 2
    assert {x.fields['name'] for x in report.discarded_objects} == {'1', 'b1'}

    obj_author = next(x for x in report.loaded_objects if x.fields['name'] == '2')
    assert report.pk_map[obj_author.id] == Author.objects.get(name='2')



'''
A nifty CLI tool to dump an exported file's contents for debugging.

Call it like this::

    python -m haul.cli /path/to/file.haul
'''

import argparse
import django
from django.conf import settings
from . import ImportContainer


def main():
    parser = argparse.ArgumentParser(description='Haul CLI')
    parser.add_argument('path', type=str)
    args = parser.parse_args()
    container = ImportContainer(ignore_unknown=True)

    settings.configure()
    django.setup()

    with open(args.path, 'rb') as f:
        with container.read(f):
            container.dump_objects()


if __name__ == '__main__':
    main()



from haul.types import ContainerFormat
from unittest import mock
from haul import ExportPolicy, ImportPolicy, RelinkAction, Attachment
from io import BytesIO
from test_app.app import reset
from test_app.models import Book, Author, Tag
from test_app.export import BookExporter, AuthorExporter, TagExporter

from haul import ImportContainer, ExportContainer


EXPORTERS = [BookExporter, AuthorExporter, TagExporter]


def test_attachments():
    reset()
    a1 = Author.objects.create(name='1')
    b1 = Book.objects.create(name='b1', author=a1)

    class EPolicy(ExportPolicy):
        def get_attachments(self, instance):
            print(instance)
            if instance == b1:
                return [Attachment.from_data('content', b'text')]
            return []

    d = BytesIO()

    c = ExportContainer(exporters=EXPORTERS, policy=EPolicy())
    c.export_objects(Book.objects.all())
    c.write(d, format=ContainerFormat.COMPRESSED_ZIP)

    reset()

    class IPolicy(ImportPolicy):
        def process_attachment(self, instance, key, stream):
            self.file_content = stream.read()

    ipolicy = IPolicy()
    ipolicy.process_attachment = mock.Mock(wraps=ipolicy.process_attachment)

    c = ImportContainer(exporters=EXPORTERS, policy=ipolicy)
    with c.read(d):
        c.import_objects()

        assert len(ipolicy.process_attachment.call_args_list) == 1
        assert ipolicy.process_attachment.call_args_list[0][0][0] == Book.objects.first()
        assert ipolicy.process_attachment.call_args_list[0][0][1] == 'content'
        assert ipolicy.file_content == b'text'



from .containers import ImportContainer, ExportContainer
from .errors import ModelClassNotRegistered
from .policy import ManyToManyImportAction, RelinkAction, BaseRelinkAction, ImportPolicy, ExportPolicy
from .serializers import Exporter, ForeignKey, ReverseForeignKey, ManyToMany
from .types import ID, Ref, Attachment, ObjectData, ContainerFormat

__all__ = [
    'ImportContainer',
    'ExportContainer',

    'ModelClassNotRegistered',

    'RelinkAction',
    'BaseRelinkAction',
    'ContainerFormat',
    'ImportPolicy',
    'ExportPolicy',
    'ManyToManyImportAction',

    'Exporter',
    'ForeignKey',
    'ReverseForeignKey',
    'ManyToMany',

    'ID',
    'Ref',
    'Attachment',
    'ObjectData',
]

__version__ = '0.0.9'



import io
import uuid
from dataclasses import dataclass, field, asdict
from django.apps import registry
from django.db.models import Model
from enum import Enum, auto
from typing import Any, Callable, Dict, List, Optional, Set, Type, BinaryIO, Union
from uuid import uuid4


class ContainerFormat(Enum):
    YAML = auto()
    COMPRESSED_ZIP = auto()
    NON_COMPRESSED_ZIP = auto()


@dataclass(frozen=True)
class ID:
    kind: str
    pk: Any

    yaml_tag = 'ID'

    def __str__(self):
        return f'{self.kind}-{self.pk}'

    @classmethod
    def from_object(cls, obj: Model) -> 'ID':
        pk = obj.pk
        if isinstance(pk, uuid.UUID):
            pk = str(pk)
        return ID(ID.kind_for_model(obj), pk)

    @staticmethod
    def kind_for_model(obj: Union[Model, Type[Model]]):
        return f'{obj._meta.app_label}:{obj._meta.model_name}'

    @staticmethod
    def model_for_kind(kind: str) -> Type[Model]:
        app_label, name, *_ = kind.split(':')
        return registry.apps.get_model(app_label, name)

    def serialize(self):
        d = asdict(self)
        if isinstance(self.pk, uuid.UUID):
            d['pk'] = str(d['pk'])
        return d

    @classmethod
    def to_yaml(cls, representer, node):
        return representer.represent_mapping('ID', node.serialize())

    @classmethod
    def deserialize(cls, s: Any) -> 'ID':
        if not isinstance(s, dict):
            raise ValueError(f'Value must be a dict, not {s}')
        return ID(**s)

    @classmethod
    def from_yaml(cls, loader, node):
        return ID(**loader.construct_mapping(node, deep=True))


@dataclass(frozen=True)
class Ref:
    ids: List[ID]
    field: str
    nullable: bool = False
    weak: bool = False


@dataclass
class Attachment:
    id: str
    key: Any
    content_provider: Optional[Callable[[], BinaryIO]] = None
    _container_stream: Optional[BinaryIO] = None

    @classmethod
    def from_path(cls, key: Any, path: str):
        return cls(
            id=str(uuid4()),
            key=key,
            content_provider=lambda: io.FileIO(path, 'rb'),
        )

    @classmethod
    def from_data(cls, key: Any, data: bytes):
        return cls(
            id=str(uuid4()),
            key=key,
            content_provider=lambda: io.BytesIO(data),  # type: ignore
        )

    @classmethod
    def to_yaml(cls, representer, node):
        data = asdict(node)
        data.pop('content_provider')
        return representer.represent_dict(data)


@dataclass
class ObjectData:
    id: ID
    serialized_data: Dict[str, Any]
    fields: Optional[Dict[str, Any]] = None
    refs: List[Ref] = field(default_factory=list)
    refers_to: Set[ID] = field(default_factory=set)
    attachments: List[Attachment] = field(default_factory=list)

    def add_reference(self, ref: Ref):
        self.refs.append(ref)
        self.refers_to |= set(ref.ids)

    def __hash__(self) -> int:
        return hash(self.id)



from haul.policy import ImportPolicy, RelinkAction
from io import BytesIO
from test_app.app import reset
from test_app.models import Book, Author, Tag
from test_app.export import BookExporter, AuthorExporter, TagExporter

from haul import ImportContainer, ExportContainer


EXPORTERS = [BookExporter, AuthorExporter, TagExporter]


def test_simple_roundtrip():
    reset()
    Author.objects.create(name='1')
    Author.objects.create(name='2')

    d = BytesIO()

    c = ExportContainer(exporters=EXPORTERS)
    c.export_objects(Author.objects.all())
    c.write(d)

    reset()

    c = ImportContainer(exporters=EXPORTERS)
    with c.read(d):
        c.import_objects()

    assert Author.objects.count() == 2
    assert Author.objects.order_by('name').first().name == '1'
    assert Author.objects.order_by('name').last().name == '2'


def test_fk():
    reset()
    a1 = Author.objects.create(name='1')
    Author.objects.create(name='2')
    Book.objects.create(name='b1', author=a1)

    d = BytesIO()

    c = ExportContainer(exporters=EXPORTERS)
    c.export_objects(Author.objects.all())
    c.export_objects(Book.objects.all())
    c.write(d)

    reset()

    c = ImportContainer(exporters=EXPORTERS)
    with c.read(d):
        c.import_objects()

    assert Author.objects.count() == 2
    assert Book.objects.count() == 1
    assert Book.objects.first().author == Author.objects.get(name='1')



import logging
import shutil
import zipfile
from django.db.models import Model
from io import TextIOWrapper
from typing import Any, Dict, IO, Iterable, List, Optional, Set, Type

from ..policy import ExportPolicy
from ..serializers import Exporter
from ..types import ID, ContainerFormat, Ref, ObjectData
from .._util import UncloseableStream, get_model_options

from .base import BaseContainer
from ._yaml import get_yaml


logger = logging.getLogger('haul.export')


class ExportContainer(BaseContainer):
    '''
    Your starting point for object export.
    '''
    __instance_map: Dict[ID, Any]

    def __init__(
        self,
        exporters: List[Type[Exporter]] = [],
        policy: Optional[ExportPolicy] = None,
        ignore_unknown=False,
    ):
        super().__init__(exporters, ignore_unknown)
        self.__instance_map = {}
        self.policy = policy or ExportPolicy()

    def export_objects(self, objects: Iterable[Model]):
        '''
        Serializes objects and adds them to the container.
        '''
        objects = list(objects)
        if not objects:
            return

        if len(set(ID.kind_for_model(obj) for obj in objects)) > 1:
            raise ValueError('Objects must be of the same class')

        outstanding_refs: Dict[str, Set[ID]] = {}

        model_meta = get_model_options(objects[0].__class__)

        kind = ID.kind_for_model(objects[0])

        objects = [
            instance for instance in objects
            if ID.from_object(instance) not in self.__instance_map and
            self.policy.should_export_object(instance)
        ]

        if not objects:
            return

        exporter_cls = self._exporter_for_model(objects[0])
        exporter = exporter_cls(objects, many=True, context={'export_policy': self.policy})

        logger.debug(f'Exporting {len(objects)} objects of kind {kind}')
        for instance, serialized_data in zip(objects, exporter.data):
            if not instance._meta.pk:
                raise ValueError(f'Cannot export models without a PK')
            pk = serialized_data.pop(instance._meta.pk.name)
            id = ID(kind, pk)
            self.__instance_map[id] = pk
            serialized_obj = ObjectData(id, serialized_data)

            # Collect references
            for key, value in list(serialized_data.items()):
                field_meta = model_meta.get_field(key)

                # Foreign key
                if isinstance(value, ID):
                    if field_meta.null:
                        if not self._should_follow_optional_reference(instance, value, key):
                            serialized_data[key] = None
                            continue

                    outstanding_refs.setdefault(value.kind, set()).add(value)
                    serialized_obj.add_reference(Ref([value], key, nullable=field_meta.null))

                # Reverse foreign key
                if isinstance(value, list) and len(value) and isinstance(value[0], ID):
                    for item in list(value):
                        if field_meta.null:
                            if not self._should_follow_optional_reference(instance, item, key):
                                value.remove(item)
                                continue
                        outstanding_refs.setdefault(item.kind, set()).add(item)
                    serialized_obj.add_reference(Ref(value, key, nullable=field_meta.null))

            # Collect attachments
            serialized_obj.attachments = self.policy.get_attachments(instance) or []

            self._objects[id] = serialized_obj

        for kind, ids in outstanding_refs.items():
            ids = ids - self._objects.keys()
            if ids:
                model_cls = ID.model_for_kind(kind)
                self.export_objects(model_cls.objects.filter(pk__in=[x.pk for x in ids]))

    def _should_follow_optional_reference(self, instance: Model, target: ID, field: str):
        if self.ignore_unknown and not self._exporter_for_kind(target.kind, raise_exception=False):
            logger.debug(f'Ignoring object {target} of unregistered kind')
            return False
        return self.policy.should_follow_reference(instance, target, field)

    def iter_objects(self) -> Iterable[ObjectData]:
        return self._objects.values()

    def write(self, stream: IO[bytes], format: ContainerFormat = ContainerFormat.YAML, metadata: Any = None):
        '''
        Writes the serialized objects from the container into a data stream.

        :param stream: The stream to write into. For ``ContainerFormat.ZIP_*``, has to be seekable.
        :param metadata: a free-form metadata object that will be stored in the stream. It's available later through :attr:`ImportContainer.metadata`.
        '''
        stream = UncloseableStream(stream)
        yaml = get_yaml()
        if format == ContainerFormat.YAML:
            for obj in self._objects.values():
                if len(obj.attachments):
                    raise ValueError('File attachments require a ZIP based container format')

            archive = None
            metadata_stream = TextIOWrapper(stream)  # type: ignore
        else:
            archive = zipfile.ZipFile(
                stream,
                mode='w',
                compression=zipfile.ZIP_STORED if format == ContainerFormat.NON_COMPRESSED_ZIP else zipfile.ZIP_DEFLATED,
            )
            metadata_stream = TextIOWrapper(archive.open('metadata.yaml', 'w'))

        try:
            try:
                header = {
                    '_': 'header',
                    'version': 1,
                    'object_kinds': list(set(x.kind for x in self._objects.keys())),
                    'metadata': metadata,
                }
                metadata_stream.write('\n---\n')
                yaml.dump(header, metadata_stream)
                for obj in self._objects.values():
                    data = {
                        '_': 'object',
                        'id': obj.id,
                        'data': obj.serialized_data,
                        'attachments': obj.attachments,
                    }
                    metadata_stream.write('\n---\n')
                    yaml.dump(data, metadata_stream)
            finally:
                metadata_stream.close()

            for obj in self._objects.values():
                for attachment in obj.attachments:
                    if not archive:
                        raise ValueError('Attachments specified not container format is not a ZIP')
                    with archive.open(f'attachments/{attachment.id}', 'w') as output:
                        assert attachment.content_provider is not None
                        attachment_stream = attachment.content_provider()
                        try:
                            shutil.copyfileobj(attachment_stream, output)  # type: ignore
                        finally:
                            attachment_stream.close()

        finally:
            if archive:
                archive.close()



from django.db.models.base import Model
from django.db.models.options import Options
from typing import Type


class UncloseableStream:
    def __init__(self, stream):
        self._stream = stream

    # Prevents ZipFile from closing the stream
    def __getattribute__(self, name: str):
        if name == 'close':
            return lambda: None
        return getattr(object.__getattribute__(self, '_stream'), name)


def get_model_options(model_cls: Type[Model]) -> Options:
    _cm = model_cls._meta.concrete_model
    if not _cm:
        raise RuntimeError(f'No concrete model for {model_cls}')
    return _cm._meta
